/*
 * Copyright (c) 1997,1998 Massachusetts Institute of Technology
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <stdio.h>
#include <math.h>

#include <fftw_mpi.h>
#include <fftw-int.h>

/* return the factor of n closest to sqrt(n): */
static int find_sqrt_factor(int n)
{
     int i = sqrt(n) + 0.5;
     int i2 = i - 1;
     
     while (i2 > 0) {
	  if (n % i2 == 0)
	       return i2;
	  if (n % i == 0)
	       return i;
	  ++i; --i2;
     }
     return 1; /* n <= 1 */
}

void fftw_mpi_local_sizes(fftw_mpi_plan p,
			  int *local_n,
			  int *local_start,
			  int *local_n_after_transform,
			  int *local_start_after_transform,
			  int *total_local_size)
{
     if (p) {
	  *local_n = p->local_m * p->r;
	  *local_start = p->local_m_start * p->r;
	  *local_n_after_transform = p->local_r * p->m;
	  *local_start_after_transform = p->local_r_start * p->m;

	  *total_local_size =
	       transpose_mpi_get_local_storage_size(p->p_transpose->nx,
						    p->p_transpose->ny,
						    p->p_transpose->my_pe,
						    p->p_transpose->n_pes);
     }
}

static int find_best_r(int n, MPI_Comm comm)
{
     int n_pes;

     MPI_Comm_size(comm, &n_pes);

     if (n % n_pes == 0) {
	  n /= n_pes;
	  if (n % n_pes == 0)
	       return (n_pes * find_sqrt_factor(n / n_pes));
	  else
	       return (n_pes * find_sqrt_factor(n));
     }
     else
	  return find_sqrt_factor(n);
}

#define MAX2(a,b) ((a) > (b) ? (a) : (b))

fftw_mpi_plan fftw_mpi_create_plan(MPI_Comm comm,
				   int n, fftw_direction dir, int flags)
{
     FFTW_TRIG_REAL twoPiOverN;
     fftw_mpi_plan p;
     int r, m, i, j;

     p = (fftw_mpi_plan) fftw_malloc(sizeof(struct fftw_mpi_plan_struct));

     i = find_best_r(n, comm);
     if (dir == FFTW_FORWARD)
	  m = n / (r = i);
     else
	  r = n / (m = i);

     p->n = n;
     p->r = r;
     p->m = m;

     flags |= FFTW_IN_PLACE;

     p->pr = fftw_create_plan(r, dir, flags);
     p->pm = fftw_create_plan(m, dir, flags);
	  
     p->p_transpose = transpose_mpi_create_plan(m, r, comm);
     p->p_transpose_inv = transpose_mpi_create_plan(r, m, comm);

     transpose_mpi_get_local_size(r,
				  p->p_transpose_inv->my_pe,
				  p->p_transpose_inv->n_pes,
				  &p->local_r,
				  &p->local_r_start);
     transpose_mpi_get_local_size(m,
				  p->p_transpose->my_pe,
				  p->p_transpose->n_pes,
				  &p->local_m,
				  &p->local_m_start);

     p->W = (fftw_complex *) fftw_malloc(sizeof(fftw_complex) *
					 p->local_r * (p->m - 1));

     twoPiOverN = FFTW_K2PI / (FFTW_TRIG_REAL) n;
     for (j = 0; j < p->local_r; ++j)
	  for (i = 1; i < p->m; ++i) {
	       int k = j * (p->m - 1) - 1 + i;
	       FFTW_TRIG_REAL
		    ij = (FFTW_TRIG_REAL) (i * (j + p->local_r_start));
	       c_re(p->W[k]) = FFTW_TRIG_COS(twoPiOverN * ij);
	       c_im(p->W[k]) = dir * FFTW_TRIG_SIN(twoPiOverN * ij);
	  }

     p->fft_work = (fftw_complex *) fftw_malloc(sizeof(fftw_complex) *
						MAX2(m, r));

     return p;
}

void fftw_mpi_destroy_plan(fftw_mpi_plan p)
{
     if (p) {
	  fftw_destroy_plan(p->pr);
	  fftw_destroy_plan(p->pm);
	  transpose_mpi_destroy_plan(p->p_transpose);
	  transpose_mpi_destroy_plan(p->p_transpose_inv);
	  fftw_free(p->W);
	  fftw_free(p->fft_work);
	  fftw_free(p);
     }
}

static void fftw_mpi_fprint_plan(FILE *f, fftw_mpi_plan p)
{
     fprintf(f, "mpi plan:\n");
     fprintf(f, "m = %d plan:\n", p->m);
     fftw_fprint_plan(f, p->pm);
     fprintf(f, "r = %d plan:\n", p->r);
     fftw_fprint_plan(f, p->pr);
}

void fftw_mpi_print_plan(fftw_mpi_plan p)
{
     fftw_mpi_fprint_plan(stdout, p);
}

void fftw_mpi(fftw_mpi_plan p, int n_fields,
	      fftw_complex *local_data, fftw_complex *work)
{
     int i, j;
     int el_size = (sizeof(fftw_complex) / sizeof(TRANSPOSE_EL_TYPE))
                   * n_fields;
     fftw_complex *W, *fft_work;
     int m;

     if (n_fields < 1)
	  return;

     transpose_mpi(p->p_transpose, el_size,
		   (TRANSPOSE_EL_TYPE *) local_data,
		   (TRANSPOSE_EL_TYPE *) work);

     W = p->W;
     m = p->m;
     fft_work = work ? work : p->fft_work;
     for (i = 0; i < p->local_r; ++i, W += m - 1) {
	  fftw_complex *d = local_data + i * (m * n_fields);

	  fftw(p->pm, n_fields, d, n_fields, 1, fft_work, 1, 0);

	  if (n_fields > 1)
	       for (j = 1; j < m; ++j) {
		    fftw_real
			 w_re = c_re(W[j-1]),
			 w_im = c_im(W[j-1]);
		    int f;
		    
		    for (f = 0; f < n_fields; ++f) {
			 fftw_real
			      d_re = c_re(d[j*n_fields + f]),
			      d_im = c_im(d[j*n_fields + f]);
			 c_re(d[j*n_fields + f]) = w_re * d_re - w_im * d_im;
			 c_im(d[j*n_fields + f]) = w_re * d_im + w_im * d_re;
		    }
	       }
	  else
               for (j = 1; j < m; ++j) {
                    fftw_real w_re = c_re(W[j-1]),
                         w_im = c_im(W[j-1]),
			 d_re = c_re(d[j]),
			 d_im = c_im(d[j]);
		    c_re(d[j]) = w_re * d_re - w_im * d_im;
		    c_im(d[j]) = w_re * d_im + w_im * d_re;
	       }
     }

     transpose_mpi(p->p_transpose_inv, el_size,
		   (TRANSPOSE_EL_TYPE *) local_data,
		   (TRANSPOSE_EL_TYPE *) work);

     if (n_fields > 1)
	  for (j = 0; j < p->local_m; ++j)
	       fftw(p->pr, n_fields, local_data + j*(p->r*n_fields),
		    n_fields, 1,
		    fft_work, 1, 0);
     else
	  fftw(p->pr, p->local_m, local_data, 1, p->r, fft_work, 1, 0);

     transpose_mpi(p->p_transpose, el_size,
		   (TRANSPOSE_EL_TYPE *) local_data,
		   (TRANSPOSE_EL_TYPE *) work);

     /* Yes, we really had to do three transposes...sigh. */
}
